import logging

import balder

from balderhub.ble.lib.scenario_features.gatt import GattBatteryServiceFeature
from balderhub.ble.lib.scenario_features.gatt.gatt_controller_feature import GattControllerFeature
from balderhub.ble.lib.scenario_features.device_information_config import DeviceInformationConfig

logger = logging.getLogger(__name__)


# TODO split up this scenario into single components to make sure that subset of tests can run even without test hw
class ScenarioBatteryService(balder.Scenario):
    """Test scenario that validates the device information service"""

    class DeviceUnderTest(balder.Device):
        """device detecting the row heart rate"""
        # _is_cnn = IsConnected()  # TODO
        config = DeviceInformationConfig()
        bat = GattBatteryServiceFeature()

    @balder.connect(DeviceUnderTest, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class Controller(balder.Device):
        """device receiving the heart rate data"""
        gatt_controller = GattControllerFeature()

    @balder.fixture('variation')
    def make_sure_that_we_are_connected(self):
        """fixture that ensures that controller is connected"""
        yield from self.Controller.gatt_controller.fixt_make_sure_to_be_connected_to(
            with_address=self.DeviceUnderTest.config.mac_address,
            restore_entry_state=True
        )

    def test_read_battery_level_once(self) -> None:
        """
        Simple test that reads the battery level once and validates that it is between 0 and 100
        """
        msg = self.Controller.gatt_controller.read(self.DeviceUnderTest.bat.BatteryLevel)
        logger.info(f'read battery level of {msg.level_percent}%')
        assert 0 <= msg.level_percent <= 100, \
            f"battery level needs to be between 0 and 100, is {msg.level_percent} (raw data: {msg.raw_data})"



import uuid
from typing import Union

from .base_gatt_service_feature import BaseGattServiceFeature
from ...utils import Characteristic, CharacteristicProperty, CharacteristicSecurity
from ...utils.base_gatt_message import BaseGattMessage


class HeartRateServiceMessage(BaseGattMessage):
    """
    GATT Message for :class:`GattHeartRateServiceFeature.MeasurementCharacteristic`
    """
    # TODO unit testen

    min_allowed_length_bytes = 2
    max_allowed_length_bytes = None


    @property
    def _flags_field(self):
        """the full flag field byte"""
        return self._raw_data[0]

    @property
    def bit_hr_format(self):
        """
        the Heart Rate Value Format bit: indicates if the data format of the heart-rate measurement value field is
        UINT8 (bit=0) or UINT16 (bit=1)
        """
        return bool(self._flags_field & (1 << 0))

    @property
    def bit_sensor_contact_status(self):
        """
        the sensor contact status bit: indicates if the sensor is attached at the moment (attached when bit is 1;
        if sensor does not detect skin contact or has poor skin contact this bit should be set to 0)
        """
        return bool(self._flags_field & (1 << 1))

    @property
    def bit_sensor_contact_support(self):
        """
        the sensor contact support bit: indicates if the sensor can detect skin contact and the value of
        `bit_sensor_contact_status` can be interpreted
        """
        return bool(self._flags_field & (1 << 2))

    @property
    def bit_energy_expended_support(self):
        """
        the energy expended support bit: indicates if the energy expended value is available within the data
        """
        return bool(self._flags_field & (1 << 3))

    @property
    def bit_rr_value_support(self):
        """
        the RR-Value support bit: indicates if the RR-Value is available within the data
        """
        return bool(self._flags_field & (1 << 4))

    @property
    def bpm(self) -> int:
        """returns the bpm value of the heart rate"""
        if self.bit_hr_format:
            return int.from_bytes(self._raw_data[1:3], byteorder='little')
        return int(self.raw_data[1])

    @property
    def sensor_has_contact(self) -> Union[bool, None]:
        """
        :return: returns True if the sensor has skin contact or False if it has no skin contact; returns None if
                 this information is not supported by the client
        """
        if self.bit_sensor_contact_support:
            return self.bit_sensor_contact_status
        return None

    @property
    def energy_expended_value(self) -> Union[int, None]:
        """represents the accumulated energy expended in kilo joules or None if no support is given"""
        if not self.bit_energy_expended_support:
            return None

        idx = 1
        idx = (idx + 2) if self.bit_hr_format else (idx + 1)
        return int.from_bytes(self.raw_data[idx:idx+2], byteorder=self.BYTE_ORDER)

    @property
    def rr_values(self) -> Union[list[int], None]:
        """returns a list of all rr-values or None if RR-Values are not supported"""
        if not self.bit_rr_value_support:
            return None

        idx = 1
        idx = (idx + 2) if self.bit_hr_format else (idx + 1)
        idx = (idx + 2) if self.bit_energy_expended_support else idx

        remaining_raw_data = self.raw_data[idx:]

        if len(remaining_raw_data) % 2 != 0:
            # length of remaining elements is not equal
            raise ValueError('can not resolve data because remaining length for RR-values is uneven')
        return [
            int.from_bytes(remaining_raw_data[i:i+2], byteorder=self.BYTE_ORDER)
            for i in range(0, len(remaining_raw_data), 2)
        ]


class GattHeartRateServiceFeature(BaseGattServiceFeature):
    """
    GATT Service Feature for Heart-Rate-Service
    """

    class MeasurementCharacteristic(Characteristic):
        """characteristic representing the Heart Rate Measurement"""
        uuid = uuid.UUID('00002a37-0000-1000-8000-00805f9b34fb')
        name = 'Heart Rate Messurement'
        optional = False
        mandatory_properties = [CharacteristicProperty.NOTIFY]
        security = CharacteristicSecurity.NONE
        message_type = HeartRateServiceMessage

    class BodySensorLocation(Characteristic):
        """characteristic representing the body sensor location"""
        uuid = uuid.UUID('00002a38-0000-1000-8000-00805f9b34fb')
        name = 'Body Sensor Location'
        optional = True
        mandatory_properties = [CharacteristicProperty.READ]
        security = CharacteristicSecurity.NONE
        message_type = HeartRateServiceMessage



import enum


class CharacteristicProperty(enum.Enum):
    """enum with all properties a characteristic can have"""
    READ = 'read'
    WRITE = 'write'
    WRITE_WITHOUT_RESPONSE = 'write-without-response'
    NOTIFY = 'notify'
    INDICATE = 'indicate'



from .base_gatt_profile_feature import BaseGattProfileFeature

class GattHeartRateProfileFeature(BaseGattProfileFeature):
    """Base GATT heart rate profile feature"""
    # TODO where do we use this feature?



import logging
import time

import balder

from balderhub.ble.lib.scenario_features import BleDeviceConfig
from balderhub.ble.lib.scenario_features.advertisement_observer_feature import AdvertisementObserverFeature
from balderhub.ble.lib.scenario_features.gatt.gatt_heart_rate_profile_feature import GattHeartRateProfileFeature
from balderhub.heart.lib.scenario_features import HeartBeatFeature
from balderhub.heart.lib.scenario_features.strap_docking_feature import StrapDockingFeature

logger = logging.getLogger(__name__)


# TODO split up this scenario into single components to make sure that subset of tests can run even without test hw
class ScenarioHrProfileAdvertisements(balder.Scenario):
    """Test scenario that validates if heart-rate sensor with active hr-profile sends Advertisements correctly"""
    # TODO we just need to restart chip before starting a test

    # TODO this needs to be configurable
    TIMEOUT_TO_AWAKE_SEC = 20
    TIMEOUT_TO_SLEEP_SEC = 60

    class Heart(balder.Device):
        """device simulating the heart beat"""
        heart = HeartBeatFeature()

    @balder.connect(Heart, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class HeartRateSensor(balder.Device):
        """device detecting the row heart rate"""
        # _is_not_cnn = IsNotConnected()  # TODO
        # _has_no_attached_strap = HasAttachedStrap()
        config = BleDeviceConfig()
        hrs = GattHeartRateProfileFeature()
        strap = StrapDockingFeature()

    @balder.connect(HeartRateSensor, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class HeartRateHost(balder.Device):
        """device receiving the heart rate data"""
        adv_observer = AdvertisementObserverFeature()

    @balder.fixture('testcase')
    def start_observer(self):
        """starts the advertisement observer"""
        self.HeartRateHost.adv_observer.start_observer()
        yield
        self.HeartRateHost.adv_observer.shutdown_observer()

    @balder.fixture('testcase')
    def make_sure_to_stop_advertising(self, start_observer):  # pylint: disable=unused-argument
        """
        fixture that ensures that no advertisements are sent (and the device is really shut down) before entering
        the test
        """
        # TODO bleak does not receive all packages!!!
        self.HeartRateSensor.strap.put_off()
        start_time = time.perf_counter()
        while (time.perf_counter() - start_time) < self.TIMEOUT_TO_SLEEP_SEC * 3:
            try:
                if self.HeartRateHost.adv_observer.wait_for_new_advertisement(
                        with_mac=self.HeartRateSensor.config.mac_address, timeout_sec=self.TIMEOUT_TO_SLEEP_SEC
                ):
                    # there are still some messages -> continue
                    continue
            except TimeoutError:
                logger.info(
                    f'received timeout: did not received any advertisements since {self.TIMEOUT_TO_SLEEP_SEC} seconds'
                )
                return
        raise TimeoutError('still received advertisements')

    @balder.parametrize("with_heart_beat", [True, False])
    def test_is_advertising_as_soon_as_strap_attached(self, with_heart_beat):  # pylint: disable=unused-argument
        """
        This test makes sure that the DUT is advertising as soon as the strap is attached
        :param with_heart_beat: test will be executed once with heart beat and another time without active heart beat
        """
        start_time = time.perf_counter()
        try:
            adv_msgs = self.HeartRateHost.adv_observer.wait_for_new_advertisement(
                with_mac=self.HeartRateSensor.config.mac_address,
                timeout_sec=self.TIMEOUT_TO_SLEEP_SEC
            )
            assert False, f"received unexpected advertisement packages `{adv_msgs}`"
        except TimeoutError:
            logger.info('received timeout - no more advertisements detected -> continue test')

        try:
            self.HeartRateSensor.strap.put_on()
            time.sleep(self.TIMEOUT_TO_AWAKE_SEC)
            # check that we receive at least one
            all_received_advertisements = self.HeartRateHost.adv_observer.filter_advertisements(
                by_mac=self.HeartRateSensor.config.mac_address,
                from_start_time=start_time
            )
            assert len(all_received_advertisements) > 1, \
                (f"does not receive any advertising messages {self.TIMEOUT_TO_AWAKE_SEC} seconds after the strap was "
                 f"attached")
        finally:
            self.HeartRateSensor.strap.put_off()

    @balder.parametrize('strap_remains_attached', [True, False])
    def test_is_advertising_when_pulse_is_gone(self, strap_remains_attached):
        """
        This test validates if the DUT is still advertising when the pulse is gone.
        :param strap_remains_attached:  test will be executed once with strap attached and once without attached strap
        """
        try:
            self.HeartRateSensor.strap.put_on()
            time.sleep(self.TIMEOUT_TO_AWAKE_SEC)
            # check that we receive at least one
            all_received_advertisements = self.HeartRateHost.adv_observer.filter_advertisements(
                by_mac=self.HeartRateSensor.config.mac_address,
            )
            assert len(all_received_advertisements) > 1, \
                (f"does not receive any advertising messages {self.TIMEOUT_TO_AWAKE_SEC} seconds after the strap was "
                 f"attached")

            self.Heart.heart.stop()
            if not strap_remains_attached:
                self.HeartRateSensor.strap.put_off()
            filter_time = time.perf_counter()
            time.sleep(self.TIMEOUT_TO_SLEEP_SEC)
            all_received_advertisements = self.HeartRateHost.adv_observer.filter_advertisements(
                by_mac=self.HeartRateSensor.config.mac_address,
                from_start_time=filter_time
            )
            assert len(all_received_advertisements) > 1, \
                (f"does not receive any advertising messages within {self.TIMEOUT_TO_SLEEP_SEC} seconds after heart "
                 f"has stopped")

            # now it is expected that server stops sending advertisements
            filter_time = time.perf_counter()
            time.sleep(self.TIMEOUT_TO_SLEEP_SEC)
            all_received_advertisements = self.HeartRateHost.adv_observer.filter_advertisements(
                by_mac=self.HeartRateSensor.config.mac_address,
                from_start_time=filter_time
            )
            assert len(all_received_advertisements) == 0, \
                f"receive some advertisement messages after {self.TIMEOUT_TO_SLEEP_SEC} seconds after heart has stopped"

        finally:
            self.HeartRateSensor.strap.put_off()



import balder
import balderhub.heart.lib.scenario_features

from balderhub.ble.lib.scenario_features import BleDeviceConfig
from balderhub.ble.lib.scenario_features.gatt import GattControllerFeature, GattHeartRateServiceFeature


class BpmValueReaderFeature(balderhub.heart.lib.scenario_features.BpmValueReaderFeature):
    """
    Setup Level feature implementation of the ``balderhub.heart.lib.scenario_features.BpmValueReaderFeature`` while
    reading the values over BLE
    """
    controller = GattControllerFeature()

    class DUT(balder.VDevice):
        """device under test specifying the HRM profile"""
        config = BleDeviceConfig()
        hrs = GattHeartRateServiceFeature()

    @property
    def timeout_to_wait_for_new_message(self):
        """
        Maximum time to wait for a new BLE message
        """
        return 3

    def prepare(self):
        self.controller.connect(self.DUT.config.mac_address, pair_on_connect=True)
        self.controller.start_notify(self.DUT.hrs.MeasurementCharacteristic)

    def read_last_bpm_value(self):

        msg = self.controller.wait_for_a_new_message_of(
            self.DUT.hrs.MeasurementCharacteristic,
            timeout=self.timeout_to_wait_for_new_message
        )
        return msg.bpm

    def cleanup(self):
        self.controller.cancel_notify(self.DUT.hrs.MeasurementCharacteristic)
        self.controller.unpair()



import uuid

from .base_gatt_service_feature import BaseGattServiceFeature
from ...utils import Characteristic, CharacteristicProperty, CharacteristicSecurity
from ...utils.base_gatt_message import BaseGattMessage


class BatteryLevelMessage(BaseGattMessage):
    """
    GATT Message for :class:`GattBatteryServiceFeature.BatteryLevel`
    """
    min_allowed_length_bytes = 1
    max_allowed_length_bytes = None

    @property
    def level_percent(self) -> int:
        """returns the level in percent (should be between 0 and 100)"""
        return int.from_bytes(self.raw_data, self.BYTE_ORDER)


class GattBatteryServiceFeature(BaseGattServiceFeature):
    """
    GATT Service Feature for Heart-Rate-Service
    """
    # TODO also implement stuff for other characteristics!!!

    class BatteryLevel(Characteristic):
        """characteristic representing the Battery Level"""
        uuid = uuid.UUID('00002a19-0000-1000-8000-00805f9b34fb')
        name = 'Battery Level'
        optional = True
        mandatory_properties = [CharacteristicProperty.READ]
        security = CharacteristicSecurity.NONE
        message_type = BatteryLevelMessage



from typing import Optional

import balder

from balderhub.ble.lib.utils.advertisement_message import AdvertisementMessage


class AdvertisementObserverFeature(balder.Feature):
    """
    Base scenario-level feature that manage an Advertisement observing device. It saves all received advertisements
    within the :class:`AdvertisementMessage` and provides them over the method
    :meth:`AdvertisementObserverFeature.filter_advertisements`.
    """

    def is_alive(self) -> bool:
        """
        This method returns True if the observer is currently alive and False otherwise. This could be a thread or
        a remote device.

        :return: True if the observer is alive, False otherwise
        """
        raise NotImplementedError

    def start_observer(self) -> None:
        """starts the observer"""
        raise NotImplementedError

    def shutdown_observer(self) -> None:
        """stops and shuts down the observer"""
        raise NotImplementedError

    def filter_advertisements(
            self,
            by_mac: Optional[str] = None,
            from_start_time: Optional[float] = None,
            till_end_time: Optional[float] = None
    ) -> list[AdvertisementMessage]:
        """
        This method returns all stored :class:`AdvertisementMessage` that were stored over all sessions made with this
        feature as long as it was active.

        :param by_mac: if given, this filters the messages for the given mac address
        :param from_start_time: if given, this filters the messages for the start time - all messages that occurred
                                after the provided timestamp will be returned (use ``time.perf_counter()``)
        :param till_end_time: if given, this filters the messages for the end time - all messages that occurred BEFORE
                              the provided timestamp will be returned (use ``time.perf_counter()``)
        :return: a list with all relevant advertisement messages
        """
        raise NotImplementedError


    def wait_for_new_advertisement(self, with_mac: Optional[str] = None, timeout_sec=10) -> AdvertisementMessage:
        """
        THis message wait for a new advertisement that occurs AFTER the message call was made. If ``with_mac`` is given,
        it will wait explicitly for a message from the specified mac address. If no message is received within
        ``timeout_sec`` the method will raise a :class:`TimeoutError`.

        .. note::
            Make sure to store these messages in internal buffers too, so it still is returned by
            :meth:`AdvertisementObserverFeature.filter_advertisements` too.

        :param with_mac: if given, the method will wait for a new message explicity from this mac address
        :param timeout_sec: maximum time to wait in seconds before raising an :class:`TimeoutError`
        :return: the new advertisement message
        """
        raise NotImplementedError



import uuid
import dataclasses

from balderhub.ble.lib.utils.base_gatt_message import BaseGattMessage
from balderhub.ble.lib.utils.raw_gatt_message import RawGattMessage
from balderhub.ble.lib.utils.characteristic_property import CharacteristicProperty
from balderhub.ble.lib.utils.characteristic_security import CharacteristicSecurity


@dataclasses.dataclass
class Characteristic:
    """Base data class for describing a BLE GATT characteristic"""
    uuid: uuid.UUID
    name: str
    optional: bool = False
    mandatory_properties: list[CharacteristicProperty] = dataclasses.field(default_factory=list)
    optional_properties: list[CharacteristicProperty] = dataclasses.field(default_factory=list)
    security: list[CharacteristicSecurity] = CharacteristicSecurity.NONE
    message_type: type[BaseGattMessage] = RawGattMessage

    def __hash__(self):
        return hash(self.uuid)

    def __repr__(self):
        return f"Characteristic<{self.name}({self.uuid})>"



from typing import Union
import logging
import platform
import time
import queue

from bleak import BleakClient, BleakScanner, BleakGATTCharacteristic, BleakCharacteristicNotFoundError, BLEDevice

from ..scenario_features.gatt.gatt_controller_feature import GattControllerFeature
from ..utils import Characteristic
from ..utils.async_manager_thread import AsyncManagerThread
from ..utils.base_gatt_message import BaseGattMessageTypeT
from ..utils.ble_device_information import BLEDeviceInformation

logger = logging.getLogger(__name__)


class BleakGattControllerFeature(GattControllerFeature):
    """bleak implementation of the feature :class:`GattControllerFeature`"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._thread: Union[AsyncManagerThread, None] = None
        self._bleak_client: Union[BleakClient, None] = None

        self._active_notifications = []
        self._queues_for_notifies_of: dict[type[Characteristic], queue.Queue] = {}
        self._stored_data_of_notifies_for = {}

        if platform.system() != 'Linux':
            # TODO provide support for this
            raise RuntimeError('the Bleak feature is not supported on your operating system')

    @property
    def adapter(self) -> Union[str, None]:
        """
        :return: the BLE adapter to use, or None if no specific adapter should be selected (auto use)
        """
        return None

    @property
    def connected_address(self) -> Union[str, None]:
        if self._bleak_client is None or not self._bleak_client.is_connected:
            return None
        return self._bleak_client.address  # TODO returns UUID on mac!!

    def create_and_start_thread(self) -> None:
        """
        This method creates and starts the internal connection management thread. This method needs to be called before
        doing anything with this feature.
        """
        if self._thread is not None:
            raise ValueError('another thread object exists -> can not create a new one')
        self._thread = AsyncManagerThread()
        self._thread.start()

    def stop_and_wait_for_thread(self, timeout: float = 10):
        """
        This method stops and cleans up the internal connection management thread. This method needs to be called to
        clean up all internal resources.
        """
        if self._thread is None:
            return
        logger.debug('stop internal async thread')
        self._thread.stop_and_wait_for_thread(timeout=timeout)
        self._thread = None

    async def _async_discover(
            self,
            limit_to_address: Union[str, None],
            timeout_discover: float
    ) -> list[BLEDevice]:
        if limit_to_address:
            device = await BleakScanner.find_device_by_address(
                limit_to_address, timeout=timeout_discover, adapter=self.adapter
            )
            devices = [] if device is None else [device]
        else:
            devices = await BleakScanner.discover(timeout=timeout_discover, adapter=self.adapter)
        return devices

    def discover(
            self,
            limit_to_address: Union[str, None] = None,
            timeout_discover: float = 20
    ) -> list[BLEDeviceInformation]:
        self._queues_for_notifies_of = {}
        self._stored_data_of_notifies_for = {}
        if self._thread is None:
            self.create_and_start_thread()

        result: list[BLEDevice] = self._thread.run_in_async_thread_and_wait_for_result(
            self._async_discover, limit_to_address, timeout_discover
        )
        if len(result) == 0:
            return []
        return [BLEDeviceInformation(dev.address, dev.name) for dev in result]

    async def _async_connect(self, mac_address: str, pair_on_connect, timeout_discover, timeout_connect):
        # TODO
        device = await BleakScanner.find_device_by_address(mac_address, timeout=timeout_discover, adapter=self.adapter)
        if device is None:
            raise ValueError(f'unable to connect with device at address {mac_address}')
        self._bleak_client = BleakClient(device, pair=pair_on_connect, timeout=timeout_connect,
                                         adapter=self.adapter)  # TODO should we save it like that?
        # TODO what is with Pair=True - we do not want this always (need to check when pairing is necessary)
        await self._bleak_client.connect()

    def connect(self, mac_address: str, pair_on_connect: bool = False, timeout_discover=20, timeout_connect=60):
        self._queues_for_notifies_of = {}
        self._stored_data_of_notifies_for = {}
        if self._thread is None:
            self.create_and_start_thread()

        self._thread.run_in_async_thread_and_wait_for_result(
            self._async_connect, mac_address, pair_on_connect, timeout_discover, timeout_connect
        )

    async def _async_unpair(self):
        if self._bleak_client is None:
            return
        await self._bleak_client.unpair()

    def unpair(self):
        self._thread.run_in_async_thread_and_wait_for_result(self._async_unpair)

    async def _async_disconnect(self):
        if self._bleak_client is None:
            return
        await self._bleak_client.disconnect()
        self._bleak_client = None

    def disconnect(self, timeout=10):
        start_time = time.perf_counter()
        self._thread.run_in_async_thread_and_wait_for_result(self._async_disconnect)
        self.stop_and_wait_for_thread(timeout=timeout - (time.perf_counter() - start_time))

        # read all queues
        for cur_characteristic in self._queues_for_notifies_of.keys():
            self.get_all_notify_messages_of(cur_characteristic)
        self._queues_for_notifies_of = {}

    async def _async_read(self, characteristic: type[Characteristic]):
        try:
            return await self._bleak_client.read_gatt_char(characteristic.uuid)
        except BleakCharacteristicNotFoundError as exc:
            raise self.CharacteristicNotFoundError from exc

    def read(self, characteristic: type[Characteristic]) -> BaseGattMessageTypeT:
        raw_data = self._thread.run_in_async_thread_and_wait_for_result(self._async_read, characteristic)
        return characteristic.message_type.from_bytearray(raw_data)

    async def _async_write(self, characteristic: type[Characteristic], value: bytearray, response: bool):
        try:
            return await self._bleak_client.write_gatt_char(characteristic.uuid, value, response)
        except BleakCharacteristicNotFoundError as exc:
            raise self.CharacteristicNotFoundError from exc

    def write(
            self,
            characteristic: type[Characteristic],
            value: bytearray
    ):
        self._thread.run_in_async_thread_and_wait_for_result(self._async_write, characteristic, value, False)

    def write_without_response(
            self,
            characteristic: type[Characteristic],
            value: bytearray
    ) -> None:
        return self._thread.run_in_async_thread_and_wait_for_result(self._async_write, characteristic, value, True)

    async def _async_start_notify(
            self,
            characteristic: type[Characteristic]
    ):
        def callback(sender, data):
            logger.info(f"received {data} from {sender}")
            # TODO what's about `sender`
            self._queues_for_notifies_of[characteristic].put((sender, data))
            # TODO send data by Queue

        await self._bleak_client.start_notify(characteristic.uuid, callback)

    def start_notify(
            self,
            characteristic: type[Characteristic]
    ):
        self._queues_for_notifies_of[characteristic] = queue.Queue()
        if characteristic not in self._stored_data_of_notifies_for:
            self._stored_data_of_notifies_for[characteristic] = []
        try:
            self._thread.run_in_async_thread_and_wait_for_result(self._async_start_notify, characteristic)
            self._active_notifications.append(characteristic)
        except BleakCharacteristicNotFoundError as exc:
            # TODO should we clean up `_stored_data_of_notifies_for` and `_queues_for_notifies_of`
            raise self.CharacteristicNotFoundError from exc

    def is_notification_active(self, characteristic: type[Characteristic]) -> bool:
        return characteristic in self._active_notifications

    async def _async_cancel_notify(self, characteristic: Characteristic):
        await self._bleak_client.stop_notify(characteristic.uuid)
        if characteristic in self._active_notifications:
            self._active_notifications.remove(characteristic)

    def cancel_notify(
            self,
            characteristic: Characteristic
    ):
        try:
            self._thread.run_in_async_thread_and_wait_for_result(self._async_cancel_notify, characteristic)
        except BleakCharacteristicNotFoundError as exc:
            # TODO should we clean up `_stored_data_of_notifies_for` and `_queues_for_notifies_of`
            raise self.CharacteristicNotFoundError from exc

    def _get_next_notify_message_from_queue(
            self,
            characteristic: type[Characteristic],
            timeout: Union[float, None] = 10
    ) -> Union[tuple[BleakGATTCharacteristic, bytearray], None]:
        if characteristic not in self._queues_for_notifies_of:
            return None
        try:
            sender, msg = self._queues_for_notifies_of[characteristic].get(block=timeout is not None, timeout=timeout)
            self._stored_data_of_notifies_for[characteristic].append((sender, msg))
            return sender, msg
        except queue.Empty:
            return None

    def get_all_notify_messages_of(
            self,
            characteristic: type[Characteristic]
    ) -> list[BaseGattMessageTypeT]:
        if characteristic not in (list(self._queues_for_notifies_of) + list(self._stored_data_of_notifies_for)):
            raise KeyError(f'no notify was registered for characteristic {characteristic}')
        # check if there are some new messages
        if characteristic in self._stored_data_of_notifies_for:
            while True:
                if not self._get_next_notify_message_from_queue(characteristic, timeout=None):
                    break
        return [
            characteristic.message_type.from_bytearray(m)
            for _, m in self._stored_data_of_notifies_for[characteristic]
        ]

    def wait_for_first_notification_of(
            self,
            characteristic: type[Characteristic],
            timeout: float = 10
    ) -> BaseGattMessageTypeT:
        start_time = time.perf_counter()
        msgs = self.get_all_notify_messages_of(characteristic)
        if len(msgs) > 0:
            return msgs[0]
        first_entry = self._get_next_notify_message_from_queue(characteristic,
                                                               timeout=timeout - (time.perf_counter() - start_time))
        if first_entry is None:
            raise TimeoutError(f'no values received within {timeout} seconds')
        return characteristic.message_type.from_bytearray(first_entry[1])

    def wait_for_a_new_message_of(
            self,
            characteristic: type[Characteristic],
            timeout: float = 10
    ) -> BaseGattMessageTypeT:
        start_time = time.perf_counter()
        # load all prevoius received messages from queue
        self.get_all_notify_messages_of(characteristic)
        first_entry = self._get_next_notify_message_from_queue(characteristic,
                                                               timeout=timeout - (time.perf_counter() - start_time))
        if first_entry is None:
            raise TimeoutError(f'no values received within {timeout} seconds')
        return characteristic.message_type.from_bytearray(first_entry[1])



from .bleak_gatt_controller_feature import BleakGattControllerFeature
from .bleak_advertisement_listener_feature import BleakAdvertisementObserverFeature


__all__ = [
    'BleakGattControllerFeature',
    'BleakAdvertisementObserverFeature',
]



from .advertisement_message import AdvertisementMessage
from .async_manager_thread import AsyncManagerThread
from .base_gatt_message import BaseGattMessage, BaseGattMessageTypeT
from .ble_device_information import BLEDeviceInformation
from .characteristic import Characteristic
from .characteristic_property import CharacteristicProperty
from .characteristic_security import CharacteristicSecurity
from .pnp_id_data import PnpIdData
from .raw_gatt_message import RawGattMessage

__all__ = [
    'AdvertisementMessage',
    'AsyncManagerThread',
    'BaseGattMessage', 'BaseGattMessageTypeT',
    'BLEDeviceInformation',
    'Characteristic',
    'CharacteristicProperty',
    'CharacteristicSecurity',
    'PnpIdData',
    'RawGattMessage',
]



from .scenario_battery_service import ScenarioBatteryService
from .scenario_device_information_service import ScenarioDeviceInformationService
from .scenario_hr_profile_advertisements import ScenarioHrProfileAdvertisements
from .scenario_validate_heart_rate_service import ScenarioHeartRateService

__all__ = [
    'ScenarioBatteryService',
    'ScenarioDeviceInformationService',
    'ScenarioHrProfileAdvertisements',
    'ScenarioHeartRateService',
]



from .base_gatt_message import BaseGattMessage


class RawGattMessage(BaseGattMessage):
    """Raw GATT message without any interpretation"""
    min_allowed_length_bytes = 0
    max_allowed_length_bytes = None



from abc import ABC, abstractmethod
from typing import Literal, TypeVar, Union


class BaseGattMessage(ABC):
    """
    Base Gatt Message class that is used for all GATT messages within this project
    """

    BYTE_ORDER: Literal["big", "little"] = 'little'

    def __init__(self, raw_data: bytearray):

        if not isinstance(raw_data, bytearray):
            raise TypeError("raw_data must be bytearray")

        if len(raw_data) < self.min_allowed_length_bytes:
            raise ValueError(f'received less bytes than expected: raw data length {len(raw_data)} < expected max '
                             f'length of {self.min_allowed_length_bytes} (raw data: `{raw_data}`)')

        if self.max_allowed_length_bytes and len(raw_data) > self.max_allowed_length_bytes:
            raise ValueError(f'received more bytes than expected: raw data length {len(raw_data)} > expected max '
                             f'length of {self.max_allowed_length_bytes} (raw data: `{raw_data}`)')

        self._raw_data = raw_data


    @property
    @abstractmethod
    def min_allowed_length_bytes(self) -> int:
        """min allowed message length in bytes (<=min_allowed_length_bytes)"""

    @property
    @abstractmethod
    def max_allowed_length_bytes(self) -> Union[int, None]:
        """max allowed message length in bytes (>=min_allowed_length_bytes) (None means no limit)"""

    @classmethod
    def from_bytearray(cls, raw_data: bytearray) -> 'BaseGattMessage':
        """
        Creates the GATT message from a bytearray

        :param raw_data: the raw bytes that describe the GATT message
        :return: a new GATT message of own type
        """
        return cls(raw_data)

    @property
    def raw_data(self) -> bytearray:
        """returns a copy of the raw byte data"""
        return self._raw_data.copy()


BaseGattMessageTypeT = TypeVar("BaseGattMessageTypeT", bound=BaseGattMessage)



from .advertisement_observer_feature import AdvertisementObserverFeature
from .ble_device_config import BleDeviceConfig
from .device_information_config import DeviceInformationConfig

__all__ = [
    'AdvertisementObserverFeature',
    'BleDeviceConfig',
    'DeviceInformationConfig',
]



from typing import Union

from balderhub.ble.lib.scenario_features.ble_device_config import BleDeviceConfig
from balderhub.ble.lib.utils.pnp_id_data import PnpIdData


class DeviceInformationConfig(BleDeviceConfig):
    """configuration feature for testing the :class:`GattDeviceInformationServiceFeature`"""

    @property
    def manufacturer_name_string(self) -> Union[bytes, None]:
        """
        expected manufacturer name that should be transmitted within the DeviceInformation-Service
        (None, if this value is not expected to exist as a characteristic within the DUT)
        """
        raise NotImplementedError()

    @property
    def model_number_string(self) -> Union[bytes, None]:
        """
        expected model number that should be transmitted within the DeviceInformation-Service
        (None, if this value is not expected to exist as a characteristic within the DUT)
        """
        raise NotImplementedError()

    @property
    def serial_number_string(self) -> Union[bytes, None]:
        """
        expected serial number that should be transmitted within the DeviceInformation-Service
        (None, if this value is not expected to exist as a characteristic within the DUT)
        """
        raise NotImplementedError()

    @property
    def hardware_revision_string(self) -> Union[bytes, None]:
        """
        expected hardware revision that should be transmitted within the DeviceInformation-Service
        (None, if this value is not expected to exist as a characteristic within the DUT)
        """
        raise NotImplementedError()

    @property
    def firmware_revision_string(self) -> Union[bytes, None]:
        """
        expected firmware revision that should be transmitted within the DeviceInformation-Service
        (None, if this value is not expected to exist as a characteristic within the DUT)
        """
        raise NotImplementedError()

    @property
    def software_revision_string(self) -> Union[bytes, None]:
        """
        expected software revision that should be transmitted within the DeviceInformation-Service
        (None, if this value is not expected to exist as a characteristic within the DUT)
        """
        raise NotImplementedError()

    @property
    def system_id(self) -> Union[bytes, None]:
        """
        expected system ID that should be transmitted within the DeviceInformation-Service
        (None, if this value is not expected to exist as a characteristic within the DUT)
        """
        raise NotImplementedError()

    @property
    def iee11073_20601_regulatory_cert_data_list(self) -> Union[bytes, None]:
        """
        expected IEE11073-20601 Regulatory Certification Data List that should be transmitted within the
        DeviceInformation-Service (None, if this value is not expected to exist as a characteristic within the DUT)
        """
        raise NotImplementedError()

    @property
    def pnp_id(self) -> Union[PnpIdData, None]:
        """
        expected PNP ID that should be transmitted within the DeviceInformation-Service
        (None, if this value is not expected to exist as a characteristic within the DUT)
        """
        raise NotImplementedError()

    @property
    def udi_for_medical(self) -> Union[bytes, None]:
        """
        expected UDI for medical devices that should be transmitted within the DeviceInformation-Service
        (None, if this value is not expected to exist as a characteristic within the DUT)
        """
        raise NotImplementedError()



import dataclasses

@dataclasses.dataclass
class BLEDeviceInformation:
    """
    provides general information about a Bluetooth Low-Energy device
    """
    #: the address of the device
    address: str
    #: the name of the device
    name: str



from .base_gatt_profile_feature import BaseGattProfileFeature
from .base_gatt_service_feature import BaseGattServiceFeature
from .gatt_battery_service_feature import GattBatteryServiceFeature
from .gatt_controller_feature import GattControllerFeature
from .gatt_device_information_service_feature import GattDeviceInformationServiceFeature
from .gatt_heart_rate_profile_feature import GattHeartRateProfileFeature
from .gatt_heart_rate_service_feature import GattHeartRateServiceFeature

__all__ = [
    "BaseGattProfileFeature",
    "BaseGattServiceFeature",
    "GattBatteryServiceFeature",
    "GattControllerFeature",
    "GattDeviceInformationServiceFeature",
    "GattHeartRateProfileFeature",
    "GattHeartRateServiceFeature",
]



import balder
import balderhub.battery.lib.scenario_features
from balderhub.ble.lib.scenario_features.ble_device_config import BleDeviceConfig
from balderhub.ble.lib.setup_features import BleakAdvertisementObserverFeature


class DeviceActivityFeature(balderhub.battery.lib.scenario_features.DeviceActivityFeature):
    """Setup Level implementation for returning the activity of an ANT device by checking if it sends messages"""

    ble_advertise = BleakAdvertisementObserverFeature()

    class DUT(balder.VDevice):
        """vdevice describing the device under test"""
        config = BleDeviceConfig()

    @property
    def timeout_to_wait_for_advertisements_sec(self) -> float:
        """
        :return: timeout to use to wait for advertisements
        """
        return 10

    def is_active(self) -> bool:
        if not self.ble_advertise.is_alive():
            self.ble_advertise.start_observer()  # TODO move in prepare/cleanup methods

        try:
            self.ble_advertise.wait_for_new_advertisement(
                self.DUT.config.mac_address,
                timeout_sec=self.timeout_to_wait_for_advertisements_sec
            )
            return True
        except TimeoutError:
            return False



from typing import Union
import time

import balder

import balderhub.ble.lib.scenario_features
import balderhub.ble.lib.scenario_features.gatt
import balderhub.ble.lib.setup_features
import balderhub.battery.lib.scenario_features


class BatteryLevelReader(balderhub.battery.lib.scenario_features.BatteryLevelReader):
    """Setup Level implementation over BLE for reading the battery level"""
    controller = balderhub.ble.lib.setup_features.BleakGattControllerFeature()

    class DUT(balder.VDevice):
        """vdevice representing the device under test that implements the ``GattBatteryServiceFeature``"""
        config = balderhub.ble.lib.scenario_features.BleDeviceConfig()
        gatt = balderhub.ble.lib.scenario_features.gatt.GattBatteryServiceFeature()

    def read_current_battery_level(self) -> Union[float, None]:

        try:
            self.controller.connect(self.DUT.config.mac_address)
            time.sleep(10)  # TODO
            return self.controller.read(self.DUT.gatt.BatteryLevel).level_percent / 100
        except ValueError as exc:
            # TODO improve
            if "unable to connect with device at address" in str(exc):
                return None
            raise exc
        finally:
            self.controller.unpair()
            #self.controller.disconnect()



import balder


class BleDeviceConfig(balder.Feature):
    """general scenario level feature for a BLE device"""

    @property
    def mac_address(self) -> str:
        """the device's MAC address"""
        raise NotImplementedError()



import struct
import dataclasses


@dataclasses.dataclass
class PnpIdData:
    """PNP ID Data object"""
    vendor_id_source: bytes
    vendor_id: bytes
    product_id: bytes
    product_version: tuple[int, int, int]

    def validate(self) -> None:
        """
        Validates the PNP ID Data object according to the requirements of the HR-Profile
        """
        if len(self.vendor_id_source) != 1:
            raise ValueError('vendor_id_source needs to be exactly one bytes1')
        if len(self.vendor_id) != 2:
            raise ValueError('vendor_id needs to be exactly two bytes')
        if len(self.product_id) != 2:
            raise ValueError('product_id needs to be exactly two bytes')
        if len(self.product_version) != 3:
            raise ValueError('product_version needs to be a tuple of exactly three numbers (major, minor, sub-minor)')

        if not isinstance(self.product_version[0], int) or not 0 <= self.product_version[0] <= 255:
            raise ValueError('major number of product_version needs to be an integer between 0 and 255')
        if not isinstance(self.product_version[1], int) or not 0 <= self.product_version[1] <= 16:
            raise ValueError('minor number of product_version needs to be an integer between 0 and 16')
        if not isinstance(self.product_version[2], int) or not 0 <= self.product_version[2] <= 16:
            raise ValueError('sub-minor number of product_version needs to be an integer between 0 and 16')

    def to_bytes(self) -> bytes:
        """
        :return: converts the dataclass into a GATT compliant bytes object
        """
        self.validate()

        version = struct.pack(
            '>BB',
            self.product_version[0],
            (self.product_version[1] << 4) | (self.product_version[2])
        )

        return version + self.product_id + self.vendor_id + self.vendor_id_source



import uuid

from .base_gatt_service_feature import BaseGattServiceFeature
from ...utils import Characteristic, CharacteristicProperty, CharacteristicSecurity


class GattDeviceInformationServiceFeature(BaseGattServiceFeature):
    """
    GATT Service Feature for Device-Information-Service
    """

    class SystemId(Characteristic):
        """characteristic representing the System ID"""

        uuid = uuid.UUID('00002a23-0000-1000-8000-00805f9b34fb')
        name = 'System ID'
        optional = True
        mandatory_properties = [CharacteristicProperty.READ]
        security = CharacteristicSecurity.NONE

    class ModelNumberCharacteristic(Characteristic):
        """characteristic representing the Model Number String"""
        uuid = uuid.UUID('00002a24-0000-1000-8000-00805f9b34fb')
        name = 'Model Number String'
        optional = True
        mandatory_properties = [CharacteristicProperty.READ]
        security = CharacteristicSecurity.NONE

    class SerialNumberCharacteristic(Characteristic):
        """characteristic representing the Serial Number"""
        uuid = uuid.UUID('00002a25-0000-1000-8000-00805f9b34fb')
        name = 'Serial Number String'
        optional = True
        mandatory_properties = [CharacteristicProperty.READ]
        security = CharacteristicSecurity.NONE

    class FirmwareRevisionCharacteristic(Characteristic):
        """characteristic representing the Firmware Revision"""
        uuid = uuid.UUID('00002a26-0000-1000-8000-00805f9b34fb')
        name = 'Firmware Revision String'
        optional = True
        mandatory_properties = [CharacteristicProperty.READ]
        security = CharacteristicSecurity.NONE

    class HardwareRevisionCharacteristic(Characteristic):
        """characteristic representing the Hardware Revision"""
        uuid = uuid.UUID('00002a27-0000-1000-8000-00805f9b34fb')
        name = 'Hardware Revision String'
        optional = True
        mandatory_properties = [CharacteristicProperty.READ]
        security = CharacteristicSecurity.NONE

    class SoftwareRevisionCharacteristic(Characteristic):
        """characteristic representing the Software Revision"""
        uuid = uuid.UUID('00002a28-0000-1000-8000-00805f9b34fb')
        name = 'Software Revision String'
        optional = True
        mandatory_properties = [CharacteristicProperty.READ]
        security = CharacteristicSecurity.NONE

    class ManufacturerCharacteristic(Characteristic):
        """characteristic representing the Manufacturer Name"""
        uuid = uuid.UUID('00002a29-0000-1000-8000-00805f9b34fb')
        name = 'Manufacturer Name String'
        optional = True
        mandatory_properties = [CharacteristicProperty.READ]
        security = CharacteristicSecurity.NONE

    class IeeeRegulatoryCertDataListCharacteristic(Characteristic):
        """characteristic representing the IEEE 11073-20601 Regulatory Certification Data List"""
        uuid = uuid.UUID('00002a2a-0000-1000-8000-00805f9b34fb')
        name = 'IEEE 11073-20601 Regulatory Certification Data List'
        optional = True
        mandatory_properties = [CharacteristicProperty.READ]
        security = CharacteristicSecurity.NONE

    class PnpIdCharacteristic(Characteristic):
        """characteristic representing the PNP ID"""
        uuid = uuid.UUID('00002a50-0000-1000-8000-00805f9b34fb')
        name = 'PnP ID'
        optional = True
        mandatory_properties = [CharacteristicProperty.READ]
        security = CharacteristicSecurity.NONE

    class UdiForMedicalDevicesCharacteristic(Characteristic):
        """characteristic representing the UDI for medical devices"""
        uuid = uuid.UUID('00002bff-0000-1000-8000-00805f9b34fb')
        name = 'UDI for Medical Devices'
        optional = True
        mandatory_properties = [CharacteristicProperty.READ]
        security = CharacteristicSecurity.NONE



import logging

import balder

from balderhub.ble.lib.scenario_features.gatt import GattDeviceInformationServiceFeature
from balderhub.ble.lib.scenario_features.gatt.gatt_controller_feature import GattControllerFeature
from balderhub.ble.lib.scenario_features.device_information_config import DeviceInformationConfig
from balderhub.ble.lib.utils import Characteristic

logger = logging.getLogger(__name__)


# TODO split up this scenario into single components to make sure that subset of tests can run even without test hw
class ScenarioDeviceInformationService(balder.Scenario):
    """Test scenario that validates the device information service"""

    class DeviceUnderTest(balder.Device):
        """device detecting the row heart rate"""
        # _is_cnn = IsConnected()  # TODO
        config = DeviceInformationConfig()
        gatt = GattDeviceInformationServiceFeature()

    @balder.connect(DeviceUnderTest, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class Controller(balder.Device):
        """device receiving the heart rate data"""
        gatt_controller = GattControllerFeature()

    @balder.fixture('variation')
    def make_sure_that_we_are_connected(self):
        """fixture that ensures that controller is connected"""
        yield from self.Controller.gatt_controller.fixt_make_sure_to_be_connected_to(
            with_address=self.DeviceUnderTest.config.mac_address
        )

    @balder.parametrize('config_and_char', [
        ('manufacturer_name_string', DeviceUnderTest.gatt.ManufacturerCharacteristic),
        ('model_number_string', DeviceUnderTest.gatt.ModelNumberCharacteristic),
        ('serial_number_string', DeviceUnderTest.gatt.SerialNumberCharacteristic),
        ('hardware_revision_string', DeviceUnderTest.gatt.HardwareRevisionCharacteristic),
        ('firmware_revision_string', DeviceUnderTest.gatt.FirmwareRevisionCharacteristic),
        ('software_revision_string', DeviceUnderTest.gatt.SoftwareRevisionCharacteristic),
        ('system_id', DeviceUnderTest.gatt.SystemId),
        ('iee11073_20601_regulatory_cert_data_list', DeviceUnderTest.gatt.IeeeRegulatoryCertDataListCharacteristic),
        ('udi_for_medical', DeviceUnderTest.gatt.UdiForMedicalDevicesCharacteristic)
    ])
    def test_value(self, config_and_char: tuple[str, type[Characteristic]]):
        """
        This test will be executed for every existing characteristic in the Device-Information Service and validates if
        it response with the correct values. In case the characteristic is marked as non-active (see properties in
        :class:`balderhub.ble.lib.scenario_features.DeviceInformationConfig`), it will check that it raises an
        exception that the characteristic was not found.

        :param config_and_char: parametrized configuration with the property name as first element and the
                                characteristic as second
        """
        config_prop_name, characteristic_type = config_and_char

        expected_val = getattr(self.DeviceUnderTest.config, config_prop_name)
        if expected_val is not None:
            logger.info(f'validate that `{config_prop_name}` is returned correctly')
            data = self.Controller.gatt_controller.read(characteristic_type).raw_data
            logger.info(f'received `{data}` (expected `{expected_val}`)')
            assert data == expected_val, f"received `{data}` for `{config_prop_name}` (expected `{expected_val}`)"

        else:
            logger.info(f'no `{config_prop_name}` characteristic expected -> check for error while reading')
            try:
                data = self.Controller.gatt_controller.read(characteristic_type).raw_data
                assert False, \
                    f'was able to receive a value `{data}` for `{config_prop_name}`, but no value was expected'
            except self.Controller.gatt_controller.CharacteristicNotFoundError as exc:
                logger.info(f'error {exc} caught - check successful')

    def test_pnp_id(self):
        """
        Special test for the PNP-ID object.

        It validates if it response with the correct values. In case the characteristic is marked as non-active
        (see property :meth:`balderhub.ble.lib.scenario_features.DeviceInformationConfig.pnp_id`),
        it will check that it raises an exception that the characteristic was not found.
        """
        expected_val = self.DeviceUnderTest.config.pnp_id
        if expected_val is not None:
            logger.info('validate that `pnp_id` is returned correctly')
            data = self.Controller.gatt_controller.read(self.DeviceUnderTest.gatt.PnpIdCharacteristic).raw_data
            logger.info(f'received `{data}` for `pnp` (expected `{expected_val}`)')
            assert data == expected_val.to_bytes(), \
                f"received `{data}` for `pnp` (expected `{expected_val}` - as bytes: {expected_val.to_bytes()})"

        else:
            logger.info('no `pnp_id` characteristic expected -> check for error while reading')
            try:
                data = self.Controller.gatt_controller.read(self.DeviceUnderTest.gatt.PnpIdCharacteristic).raw_data
                assert False, f'was able to receive a value `{data}` for `pnp`, but no value was expected'
            except self.Controller.gatt_controller.CharacteristicNotFoundError as exc:
                logger.info(f'error {exc} caught - check successful')



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

sys.path.insert(0, os.path.abspath('../../src'))

from balderhub.ble import __version__

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'BalderHub BLE'
version = __version__
copyright = "2026, Max Stahlschmidt"
author = 'Max Stahlschmidt'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc']

templates_path = ['_templates']
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

# -- Autodoc settings --------------------------------------------------------

autodoc_default_options = {
    'show-inheritance': True
}



from typing import Union, Any

import asyncio
import threading
import time

import logging

logger = logging.getLogger(__name__)


class AsyncManagerThread:
    """
    Base AsyncManagerThread that is used to manage the async Bleak library within a sync context that is required for
    Balder.

    This thread can be created multiple times within one instance of this object. Nevertheless it needs to be shut down
    before it can be started again.
    """

    def __init__(self):
        self._async_loop = asyncio.new_event_loop()
        self._thread: Union[threading.Thread, None] = None

    def is_alive(self):
        """
        :return: returns true if the thread is alive, False otherwise
        """
        return self._thread.is_alive()

    def __threaded_method(self):
        asyncio.set_event_loop(self._async_loop)
        self._async_loop.run_forever()

    def start(self) -> None:
        """
        This method starts the async manager thread. It raises an ValueError in case the thread is already running.
        """
        if self._thread is not None:
            raise ValueError('thread was already started before -> can not restart it')
        self._thread = threading.Thread(target=self.__threaded_method, daemon=True)  # TODO should we use deamon??
        self._thread.start()

    def run_in_async_thread_and_wait_for_result(self, async_callback, *args) -> Any:
        """
        Helper method to run any async callback within the special async manager thread
        :param async_callback: the async callback to run
        :param args: all arguments to pass to the async callback
        :return: the result of the async callback
        """
        async def execute():
            await asyncio.sleep(0.01)
            result = await async_callback(*args)
            await asyncio.sleep(0.01)
            return result

        future = asyncio.run_coroutine_threadsafe(execute(), self._async_loop)
        return future.result()

    def _run_stop_command(self, timout_sec: float = 10) -> None:
        start_time = time.perf_counter()
        async def execute():
            await asyncio.sleep(0.01)
            tasks = [t for t in asyncio.all_tasks(self._async_loop) if t is not asyncio.current_task()]
            logger.debug(f'found {len(tasks)} active tasks -> cancel them')
            for task in tasks:
                task.cancel()
            await asyncio.gather(*tasks, return_exceptions=True)
            logger.debug('all active tasks canceled')
            return None

        future = asyncio.run_coroutine_threadsafe(execute(), self._async_loop)
        future.result()
        logger.debug('shutdown finished')
        logger.debug('stop loop from main thread too')
        self._async_loop.call_soon_threadsafe(self._async_loop.stop)
        logger.debug('loop from main thread stopped')

        # wait for loop
        while self._async_loop.is_running():
            if time.perf_counter() - start_time > timout_sec:
                raise TimeoutError('was unable to stop async-thread')
            time.sleep(0.01)
        self._async_loop.close()

    def stop_and_wait_for_thread(self, timeout: float = 10) -> None:
        """
        This method stops the async manager thread and waits for its termination.

        :param timeout: maximum time in seconds to wait for thread
        """
        start_time = time.perf_counter()
        self._run_stop_command()
        self._thread.join(timeout=timeout - (time.perf_counter() - start_time))
        if self._thread.is_alive():
            raise TimeoutError('was unable to stop async-thread')



import balder

from balderhub.ble.lib.utils import Characteristic


class BaseGattServiceFeature(balder.Feature):
    """
    Base abstract GATT service feature which is used as base class for every GATT service feature within this project
    """

    @classmethod
    def get_expected_characteristics(cls) -> list[type[Characteristic]]:
        """
        Determines the expected characteristics of the GATT service

        .. note::
            This is a helper method to auto determine the characteristics based on inner classes that inherit from
            :class:`Characteristic`.

        :return: returns a list of expected characteristics this service feature should handle
        """
        result = []

        for base in cls.__mro__:
            for _, val in base.__dict__.items():
                # it is a subclass of `Characteristic` and is defined within the current module
                if isinstance(val, type) and issubclass(val, Characteristic) and val.__module__ == base.__module__:
                    # make sure that it is really a nested class
                    if '.' in val.__qualname__:
                        if val not in result:
                            result.append(val)

        return result



import dataclasses
from typing import Union

from balderhub.ble.lib.utils.ble_device_information import BLEDeviceInformation


@dataclasses.dataclass
class AdvertisementMessage:
    """Base data class for BLE advertisement messages"""
    timestamp: float
    device: BLEDeviceInformation
    local_name: Union[str, None]
    manufacturer_data: dict[int, bytes]
    rssi: int
    service_data: dict[str, bytes]
    service_uuids: list[str]



from typing import Union, Generator
import logging

import balder

from balderhub.ble.lib.utils import Characteristic
from balderhub.ble.lib.utils.base_gatt_message import BaseGattMessageTypeT
from balderhub.ble.lib.utils.ble_device_information import BLEDeviceInformation

logger = logging.getLogger(__name__)


class GattControllerFeature(balder.Feature):
    """
    Universal GATT controller feature for interacting with BLE devices that implement one or more GATT
    services/profiles.
    """

    class CharacteristicNotFoundError(Exception):
        """used when the provided characteristic was not found at the connected device"""

    def discover(
            self,
            limit_to_address: Union[str, None] = None,
            timeout_discover: float = 20
    ) -> list[BLEDeviceInformation]:
        """
        Discover available devices that are in Advertising mode at the moment

        :param limit_to_address: if given, the discover procedure is limited to the provided address
        :param timeout_discover: the maximum time to discover devices (timeout, if `limit_to_address` is given)
        :return: the discovered devices
        """
        raise NotImplementedError

    def connect(
            self,
            mac_address: str,
            pair_on_connect: bool = False,
            timeout_discover=20,
            timeout_connect=60
    ) -> None:
        """
        This method connects the controller to the device with the given mac address.

        As soon as a connection is established, this feature is bound to the connected device. You can reconnect with
        another device after running the method :meth:`GattControllerFeature.disconnect`.

        :param mac_address: the mac address of the device to connect with
        :param pair_on_connect: True if the device should be paired directly while connecting
        :param timeout_discover: timeout to discover the device
        :param timeout_connect: timeout till the connection should be established
        """
        raise NotImplementedError

    def unpair(self) -> None:
        """
        Unpairs the current connected device.
        """
        raise NotImplementedError

    def disconnect(self, timeout=10) -> None:
        """
        Executed the disconnect procedure with the connected device.

        :param timeout: max time till the device needs to be disconnected
        """
        raise NotImplementedError

    @property
    def connected_address(self) -> Union[str, None]:
        """
        :return: returns the current connected device's mac address or None if this feature is not connected with a
                 device yet
        """
        raise NotImplementedError

    @property
    def is_connected(self) -> bool:
        """
        :return: returns True if the feature is connected with a device, otherwise False
        """
        return self.connected_address is not None

    def read(self, characteristic: type[Characteristic]) -> BaseGattMessageTypeT:
        """
        This method reads the value of a specific characteristic. It returns the related message.

        :param characteristic: the characteristic to read
        :return: the response for the given characteristic
        """
        raise NotImplementedError

    def write(self, characteristic: type[Characteristic], value: bytearray) -> BaseGattMessageTypeT:
        """
        This method writes the value of a specific characteristic. It returns the related response.

        :param characteristic: the characteristic to write
        :param value: the value to write
        :return: the response from the device
        """
        raise NotImplementedError

    def write_without_response(self, characteristic: type[Characteristic], value: bytearray) -> None:
        """
        This method writes the value of a specific characteristic. It does not return anything, because the method uses
        the write-without-response method.

        :param characteristic: the characteristic to write
        :param value: the value to write
        """
        raise NotImplementedError

    def start_notify(self, characteristic: type[Characteristic]) -> None:
        """
        This method starts the notification for a specific characteristic.

        :param characteristic: the characteristic the notification should be started for
        """
        raise NotImplementedError

    def cancel_notify(self, characteristic: type[Characteristic]) -> None:
        """
        This method cancels the nofify for a specific characteristic.

        :param characteristic: the characteristic the notification should be canceled for
        """
        raise NotImplementedError

    def is_notification_active(self, characteristic: type[Characteristic]) -> bool:
        """
        Checks if the notification is active for the given characteristic.

        :param characteristic: the characteristic that should be checked
        :return: True if there was a nofification registered, False otherwise
        """
        raise NotImplementedError

    # TODO
    # def start_indicate(self, characteristic: Characteristic):
    #     pass

    # def stop_indicate(self, characteristic: Characteristic):
    #     pass

    def get_all_notify_messages_of(
            self,
            characteristic: type[Characteristic]
    ) -> list[BaseGattMessageTypeT]:
        """
        This method returns all received notification messages for a given characteristic.
        :param characteristic: the characteristic the notification needs to be from
        :return: a collection with all messages ordered by timestamp
        """
        raise NotImplementedError

    def wait_for_first_notification_of(
            self,
            characteristic: type[Characteristic],
            timeout: float = 10
    ) -> BaseGattMessageTypeT:
        """
        This method waits for the first notification of the given characteristic and returns it. If a message was
        already received before this method was called, it will be returned.

        If there is no notification received within the given timeout, the method raises a TimeoutError.

        :param characteristic: the characteristic the notification needs to be from
        :param timeout: the maximum time in seconds to block, if this timeout is reached the method raises a
                        TimeoutError
        :return: the newly received GATT message
        """
        raise NotImplementedError

    def wait_for_a_new_message_of(
            self,
            characteristic: type[Characteristic],
            timeout: float = 10
    ) -> BaseGattMessageTypeT:
        """
        This method waits for a specific notification and returns as soon as it detects a new one. All buffered
        notification that were received before entering this method are ignored.

        If there is no notification received within the given timeout, the method raises a TimeoutError.

        :param characteristic: the characterisitic the notification needs to be from
        :param timeout: the maximum time in seconds to block, if this timeout is reached the method raises a
                        TimeoutError
        :return: the newly received GATT message
        """
        raise NotImplementedError

    def _fixt_teardown(self, with_address: str, controller_was_connected: bool):
        if controller_was_connected:
            if self.is_connected:
                logger.info('BLE Controller was connected before entering this fixture and it is still connected '
                            '-> do nothing')
            else:
                logger.info('BLE Controller was connected before entering this fixture, but it is not connected now '
                            '-> connect it again')
                self.connect(with_address)
        else:
            # controller was connected before this fixture
            if self.is_connected:
                logger.info('BLE Controller was disconnected before entering this fixture but it is connected now '
                            '-> disconnect it again')
                self.disconnect()
            else:
                logger.info('BLE Controller was disconnected before entering this fixture and it is still disconnected '
                            '-> do nothing')

    def fixt_make_sure_to_be_connected_to(
            self,
            with_address: str,
            restore_entry_state: bool = True
    ) -> Generator[None, None, None]:
        """
        Fixture that makes sure that the BLE connection is established with the given address. If there is another
        connection established, the method will disconnect it and reconnect with the given address.

        .. note::

            You can use this fixture directly by using:

            .. code-block:: python

                @balder.fixture(...)
                def my_fixture(...):
                    yield from feat.fixt_make_sure_to_be_connected_to(...)

        :param with_address: the address to connect with
        :param restore_entry_state: True if the previous state (either connected or disconnected) should be
                                    reestablished in the teardown part of this fixture
        :return: the ready to use fixture generator, but without any values
        """
        controller_was_connected = self.is_connected

        reconnect_afterwards_to = with_address
        if controller_was_connected:
            connected_to_mac = self.connected_address
            if with_address == connected_to_mac:
                logger.info(f'BLE Controller is already connected to {with_address} -> do nothing')
            else:
                logger.info(f'BLE Controller is connected to another device {connected_to_mac} -> disconnect it and '
                            f'connect with {with_address}')
                reconnect_afterwards_to = self.connected_address
                self.disconnect()
                self.connect(with_address)
        else:
            logger.info('BLE Controller is not connected yet -> connect it')
            self.connect(with_address)

        yield None

        if not restore_entry_state:
            return
        self._fixt_teardown(with_address=reconnect_afterwards_to, controller_was_connected=controller_was_connected)

    def fixt_make_sure_to_be_disconnected_from(
            self,
            with_address: str,
            restore_entry_state: bool = True
    ) -> Generator[None, None, None]:
        """
        Fixture that makes sure that the BLE connection is disconnected before finishing the construction part of this
        fixture. If there is another connection established to another device like the given address, the fixture will
        do nothing. If it is connected to the device with the given address it will disconnect before finishing the
        construction part.
        If `restore_entry_state` is True, the method will reestablish the state that was given before entering this
        fixture.

        .. note::

            You can use this fixture directly by using:

            .. code-block:: python

                @balder.fixture(...)
                def my_fixture(...):
                    yield from feat.fixt_make_sure_to_be_disconnected_from(...)

        :param with_address: the address the connection should be disconnected from
        :param restore_entry_state: True if the previous state (either connected or disconnected) should be
                                    reestablished in the teardown part of this fixture
        :return: the ready to use fixture generator, but without any values
        """
        controller_was_connected = self.is_connected
        should_reconnect_afterwards = controller_was_connected
        if controller_was_connected:
            connected_to_mac = self.connected_address
            if with_address == connected_to_mac:
                logger.info(f'BLE Controller is already connected to {with_address} -> disconnect it')
                self.disconnect()
            else:
                logger.info(f'BLE Controller is connected to another device {connected_to_mac} -> do nothing')
                should_reconnect_afterwards = False
        else:
            logger.info('BLE Controller is already disconnected -> do nothing')

        yield None

        if not restore_entry_state:
            return
        self._fixt_teardown(with_address=with_address, controller_was_connected=should_reconnect_afterwards)



from __future__ import annotations
import logging
import time

import balder

from balderhub.ble.lib.scenario_features import DeviceInformationConfig
from balderhub.ble.lib.scenario_features.gatt.gatt_controller_feature import GattControllerFeature
from balderhub.ble.lib.scenario_features.gatt.gatt_heart_rate_service_feature import GattHeartRateServiceFeature, \
    HeartRateServiceMessage
from balderhub.heart.lib.scenario_features import HeartBeatFeature, StrapDockingFeature

logger = logging.getLogger(__name__)


# TODO split up this scenario into single components to make sure that subset of tests can run even without test hw
class ScenarioHeartRateService(balder.Scenario):
    """Test scenario that validates a heart rate sensor according the definitions made from HeartRateService"""

    class Heart(balder.Device):
        """device simulating the heart beat"""
        heart = HeartBeatFeature()

    @balder.connect(Heart, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class HeartRateSensor(balder.Device):
        """device detecting the row heart rate"""
        # _is_cnn = IsConnected()  # TODO
        config = DeviceInformationConfig()
        hrs = GattHeartRateServiceFeature()
        strap = StrapDockingFeature()

    @balder.connect(HeartRateSensor, over_connection=balder.Connection)  # pylint: disable=undefined-variable
    class HeartRateHost(balder.Device):
        """device receiving the heart rate data"""
        gatt_controller = GattControllerFeature()

    @property
    def active_characteristic(self):
        """
        :return: helper property for getting the Measurement characteristic
        """
        return self.HeartRateSensor.hrs.MeasurementCharacteristic

    @balder.fixture('variation')
    def make_sure_that_we_are_connected(self):
        """fixture that ensures that controller is connected"""
        yield from self.HeartRateHost.gatt_controller.fixt_make_sure_to_be_connected_to(
            with_address=self.HeartRateSensor.config.mac_address,
            restore_entry_state=True
        )

    @balder.fixture('testcase')
    def start_notification(self):
        """
        :return: register the notification for the measurement characteristic
        """
        self.HeartRateHost.gatt_controller.start_notify(self.active_characteristic)
        yield
        self.HeartRateHost.gatt_controller.cancel_notify(self.active_characteristic)

    def test_validate_heart_rate(self):
        """
        Simple test that validates the heart beats sent by the sensor
        """
        self.Heart.heart.start(60)

        logger.info("wait to get first data package")
        first_msg: HeartRateServiceMessage = self.HeartRateHost.gatt_controller.wait_for_first_notification_of(
            self.active_characteristic, timeout=10
        )
        logger.info(f"received data {first_msg.bpm}")
        assert first_msg.bpm == 60, f"received {first_msg.bpm} bpm"

    def test_sensorcontact_bit(self):
        """test that validates that the sensor contact support and status bit are handled correctly"""
        self.Heart.heart.start(60)

        first_msg: HeartRateServiceMessage = self.HeartRateHost.gatt_controller.wait_for_first_notification_of(
            self.active_characteristic
        )
        if not first_msg.bit_sensor_contact_support:
            assert first_msg.bit_sensor_contact_status is False
            return

        assert first_msg.bit_sensor_contact_status is True

        self.HeartRateSensor.strap.put_off()
        time.sleep(2)
        next_msg: HeartRateServiceMessage = self.HeartRateHost.gatt_controller.wait_for_a_new_message_of(
            self.active_characteristic
        )

        assert next_msg.bit_sensor_contact_support is True
        assert next_msg.bit_sensor_contact_status is False

        self.HeartRateSensor.strap.put_on()
        time.sleep(2)
        next_msg: HeartRateServiceMessage = self.HeartRateHost.gatt_controller.wait_for_a_new_message_of(
            self.active_characteristic
        )

        assert next_msg.bit_sensor_contact_support is True
        assert next_msg.bit_sensor_contact_status is True



import platform
import time
from typing import Union, Optional
import logging

import queue

from bleak import BleakScanner, AdvertisementData, BLEDevice

import balderhub.ble.lib.scenario_features
from balderhub.ble.lib.utils.advertisement_message import AdvertisementMessage
from balderhub.ble.lib.utils.ble_device_information import BLEDeviceInformation
from balderhub.ble.lib.utils.async_manager_thread import AsyncManagerThread

logger = logging.getLogger(__name__)


class BleakAdvertisementObserverFeature(balderhub.ble.lib.scenario_features.AdvertisementObserverFeature):
    """
    Feature implementation with bleak for :class:`AdvertisementObserverFeature`

    .. note::
        This feature uses bleak as backend. Bleak does not receive every Advertisement. Consider that in mind, when
        using this feature.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._start_time = None

        self._advertisement_queue = queue.Queue()

        self._already_read_advertisements: list[AdvertisementMessage] = []

        self._thread: Union[AsyncManagerThread, None] = None

        self._bleak_scanner: Union[BleakScanner, None] = None

        if platform.system() != 'Linux':
            # TODO provide support for this
            raise RuntimeError('the Bleak feature is not supported on your operating system')

    @property
    def adapter(self) -> Union[str, None]:
        """
        :return: the BLE adapter to use, or None if no specific adapter should be selected (auto use)
        """
        return None

    async def _detection_callback(self, device, adv_data):
        logger.debug(f'received a new advertisement {adv_data} from device {device}')
        self._advertisement_queue.put((time.perf_counter(), device, adv_data))

    async def _async_create_bleak_scanner(self):
        self._bleak_scanner = BleakScanner(
            detection_callback=self._detection_callback,  # TODO should we save it like that?
            adapter=self.adapter
        )
        await self._bleak_scanner.start()
        self._start_time = time.perf_counter()

    async def _async_stop_scanner(self):
        await self._bleak_scanner.stop()

    def start_observer(self):
        if self._thread is not None:
            raise ValueError('observer is already active, can not start it again')
        self._already_read_advertisements = []
        self._thread = AsyncManagerThread()
        self._thread.start()

        self._thread.run_in_async_thread_and_wait_for_result(
            self._async_create_bleak_scanner
        )

    def is_alive(self):
        return self._thread is not None and self._thread.is_alive()

    def shutdown_observer(self, timeout=10):
        if self._thread is None:
            return

        self._thread.run_in_async_thread_and_wait_for_result(
            self._async_stop_scanner
        )

        logger.debug('stop internal async thread')
        self._thread.stop_and_wait_for_thread(timeout=timeout)
        self._thread = None

    def _read_all_new_advertisement_from_queue(self) -> list[AdvertisementMessage]:
        new_ones = []
        while not self._advertisement_queue.empty():
            data: tuple[float, BLEDevice, AdvertisementData] = self._advertisement_queue.get(block=False, timeout=None)
            timestamp, device, msg = data

            device = BLEDeviceInformation(
                address=device.address,
                name=device.name
            )

            new_ones.append(
                AdvertisementMessage(
                    timestamp=timestamp,
                    device=device,
                    local_name=msg.local_name,
                    manufacturer_data=msg.manufacturer_data,
                    rssi=msg.rssi,
                    service_data=msg.service_data,
                    service_uuids=msg.service_uuids
                )
            )
        self._already_read_advertisements.extend(new_ones)
        return new_ones

    def wait_for_new_advertisement(self, with_mac: Optional[str] = None, timeout_sec=10):
        start_time = time.perf_counter()
        while (time.perf_counter() - start_time) < timeout_sec:
            new_messages = self._read_all_new_advertisement_from_queue()
            if len(new_messages) > 0:

                filtered_msgs = new_messages \
                    if with_mac is None \
                    else [msg for msg in new_messages if msg.device.address == with_mac]

                if len(filtered_msgs) > 0:
                    return filtered_msgs[0]
            else:
                time.sleep(0.01)

        raise TimeoutError(f'no new advertisement received within {timeout_sec} seconds')

    def __validate_time(self, value: float, variable_name: str):
        if not isinstance(value, float):
            raise TypeError(f'time value `{variable_name}` needs to be an float')
        if value < self._start_time:
            raise ValueError(f'{variable_name} is less than or equal to start time - can not be the case')
        if value > time.perf_counter():
            raise ValueError('{variable_name} is in the future - can not be the case')

    def filter_advertisements(
            self,
            by_mac: Optional[str] = None,
            from_start_time: Optional[float] = None,
            till_end_time: Optional[float] = None
    ) -> list[AdvertisementMessage]:
        # read all advertisements from queu
        self._read_all_new_advertisement_from_queue()

        all_msgs = self._already_read_advertisements
        if by_mac is not None:
            all_msgs = [msg for msg in all_msgs if msg.device.address == by_mac]

        if from_start_time is not None:
            self.__validate_time(from_start_time, 'from_start_time')
            all_msgs = [msg for msg in all_msgs if msg.timestamp >= from_start_time]

        if till_end_time is not None:
            self.__validate_time(till_end_time, 'till_end_time')

            if from_start_time:
                if from_start_time > till_end_time:
                    raise ValueError('start time can not be higher than end time')

            all_msgs = [msg for msg in all_msgs if msg.timestamp <= till_end_time]

        return all_msgs



import time

import balder
import balderhub.heart.lib.scenario_features

from balderhub.ble.lib.scenario_features import BleDeviceConfig
from balderhub.ble.lib.scenario_features.gatt import GattControllerFeature, GattHeartRateServiceFeature


class RRValueReaderFeature(balderhub.heart.lib.scenario_features.RRValueReaderFeature):
    """
    Setup Level feature implementation of the ``balderhub.heart.lib.scenario_features.RRValueReaderFeature`` while
    reading the values over BLE
    """
    controller = GattControllerFeature()

    class DUT(balder.VDevice):
        """device under test specifying the HRM profile"""
        config = BleDeviceConfig()
        hrs = GattHeartRateServiceFeature()

    @property
    def timeout_to_wait_for_new_rr_value_sec(self):
        """maximum time to wait for a RR value"""
        return 3

    def prepare(self):
        self.controller.connect(self.DUT.config.mac_address, pair_on_connect=True)
        self.controller.start_notify(self.DUT.hrs.MeasurementCharacteristic)

    def read_last_rr_value_in_sec(self) -> float:
        start_time = time.perf_counter()
        while time.perf_counter() - start_time < self.timeout_to_wait_for_new_rr_value_sec:
            msg = self.controller.wait_for_a_new_message_of(
                self.DUT.hrs.MeasurementCharacteristic,
                timeout=self.timeout_to_wait_for_new_rr_value_sec - (time.perf_counter() - start_time)
            )
            if msg.rr_values:
                return msg.rr_values[-1]
        raise TimeoutError(f'did not received more than one heart beat messages with an RR-Value within '
                           f'{self.timeout_to_wait_for_new_rr_value_sec} seconds')

    def cleanup(self):
        self.controller.cancel_notify(self.DUT.hrs.MeasurementCharacteristic)
        self.controller.unpair()
